from video_transformers.auto.backbone import AutoBackbone
from video_transformers.auto.head import AutoHead
from video_transformers.auto.neck import AutoNeck
from video_transformers.modules import TimeDistributed, VideoClassificationModel

__version__ = "0.0.4"
